package org.jastadd.ragconnect.tests;

import defaultOnlyRead.ast.MqttHandler;
import org.junit.jupiter.api.*;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

/**
 * Base class for tests ensuring running mqtt broker.
 *
 * @author rschoene - Initial contribution
 */
@Tag("mqtt")
public abstract class AbstractMqttTest {

  static boolean checkDone = false;
  static boolean checkResult;
  static MqttHandler publisher;

  @BeforeAll
  public static void createPublishAndOnceCheckMqttConnection() {
    try {
      publisher = new MqttHandler("Publisher")
          .dontSendWelcomeMessage()
          .setHost(TestUtils.getMqttHost());
      checkResult = true;
    } catch (IOException e) {
      checkResult = false;
    }
    if (!checkDone) {
      checkDone = true;
      checkResult &= publisher.waitUntilReady(2, TimeUnit.SECONDS);
    }
    if (!checkResult) {
      throw new IllegalStateException("Mqtt Broker not ready!");
    }
  }

  @Test
  public final void buildModel() {
    createModel();
  }

  @Tag("mqtt")
  @Test
  public final void testCommunicateSendInitialValue() throws IOException, InterruptedException {
    createModel();
    setupReceiverAndConnect(true);

    communicateSendInitialValue();
  }

  /**
   * Actual test code for communication when sending initial value.
   * @throws InterruptedException because of TestUtils.waitForMqtt()
   */
  protected abstract void communicateSendInitialValue() throws InterruptedException;

  @Tag("mqtt")
  @Test
  public final void testCommunicateOnlyUpdatedValue() throws IOException, InterruptedException {
    createModel();
    setupReceiverAndConnect(false);

    communicateOnlyUpdatedValue();
  }

  /**
   * Actual test code for communication without sending any value upon connecting.
   * @throws InterruptedException because of TestUtils.waitForMqtt()
   */
  protected abstract void communicateOnlyUpdatedValue() throws InterruptedException;

  /**
   * Create the model, and set required default values.
   */
  protected abstract void createModel();

  /**
   * Begin with this snippet
   * <pre>
   *     model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);
   *
   *     handler = new MqttHandler().dontSendWelcomeMessage().setHost(TestUtils.getMqttHost());
   *     assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS));
   * </pre>
   *
   * And then add dependencies, initialise receiver, add connections to those receivers,
   * and finally call generated connect* methods on model elements.
   * @param writeCurrentValue if the initial/current value shall be sent upon connecting
   */
  protected abstract void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException;

  @AfterEach
  public void alwaysCloseConnections() {
    closeConnections();
  }

  /**
   * Write the following snippet (using your correct handler and model):
   * <pre>
   * if (handler != null) {
   *    handler.close();
   * }
   * if (model != null) {
   *   model.ragconnectCloseConnections();
   * }
   * </pre>
   */
  protected abstract void closeConnections();

  @AfterAll
  public static void closePublisher() {
    if (publisher != null) {
      publisher.close();
    }
  }

}
