package org.jastadd.ragconnect.tests;

import incremental.ast.A;
import incremental.ast.B;
import incremental.ast.MqttHandler;
import org.junit.jupiter.api.Tag;

import java.io.IOException;
import java.util.concurrent.TimeUnit;

import static org.jastadd.ragconnect.tests.TestUtils.mqttUri;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

/**
 * Testcase "Incremental Dependency".
 *
 * @author rschoene - Initial contribution
 */
@Tag("Incremental")
public class IncrementalDependencyTest extends AbstractMqttTest {

  private static final String TOPIC_IN = "in/a";
  private static final String TOPIC_OUT_A = "out/a";
  private static final String TOPIC_OUT_B1 = "out/b1";
  private static final String TOPIC_OUT_B2 = "out/b2";

  private MqttHandler handler;
  private A model;
  private B b1;
  private B b2;

  private ReceiverData dataA;
  private ReceiverData dataB1;
  private ReceiverData dataB2;

  @Override
  protected void createModel() {
    model = new A();
    model.setInput("Start");
    b1 = new B();
    b2 = new B();
    model.addB(b1);
    model.addB(b2);
    model.ragconnectCheckIncremental();
  }

  @Override
  protected void setupReceiverAndConnect(boolean writeCurrentValue) throws IOException {
    model.ragconnectSetupMqttWaitUntilReady(2, TimeUnit.SECONDS);

    handler = new MqttHandler("TestHandler")
        .dontSendWelcomeMessage()
        .setHost(TestUtils.getMqttHost());
    assertTrue(handler.waitUntilReady(2, TimeUnit.SECONDS));

    // no dependencies for the model are set here

    dataA = new ReceiverData();
    dataB1 = new ReceiverData();
    dataB2 = new ReceiverData();

    handler.newConnection(TOPIC_OUT_A, bytes -> {
      dataA.numberOfStringValues += 1;
      dataA.lastStringValue = TestUtils.DefaultMappings.BytesToString(bytes);
    });
    handler.newConnection(TOPIC_OUT_B1, bytes -> {
      dataB1.numberOfStringValues += 1;
      dataB1.lastStringValue = TestUtils.DefaultMappings.BytesToString(bytes);
    });
    handler.newConnection(TOPIC_OUT_B2, bytes -> {
      dataB2.numberOfStringValues += 1;
      dataB2.lastStringValue = TestUtils.DefaultMappings.BytesToString(bytes);
    });

    model.connectInput(mqttUri(TOPIC_IN));
    model.connectOutputOnA(mqttUri(TOPIC_OUT_A), writeCurrentValue);
    b1.connectOutputOnB(mqttUri(TOPIC_OUT_B1), writeCurrentValue);
    b2.connectOutputOnB(mqttUri(TOPIC_OUT_B2), writeCurrentValue);
  }

  @Override
  protected void communicateSendInitialValue() throws InterruptedException {
    // check initial value
    TestUtils.waitForMqtt();
    checkData(1, "aStart",
        "bStartPostfix",
        "bStartPostfix");

    // send and check new value
    sendData("101");
    checkData(2, "a101",
        "b101Postfix",
        "b101Postfix");

    // send and check same value
    sendData("101");
    checkData(2, "a101",
        "b101Postfix",
        "b101Postfix");

    // send and check new value
    sendData("201");
    checkData(3, "a201",
        "b201Postfix",
        "b201Postfix");
  }

  @Override
  protected void communicateOnlyUpdatedValue() throws InterruptedException {
    // check initial value
    TestUtils.waitForMqtt();
    checkData(0, null,
        null,
        null);

    // send and check new value
    sendData("102");
    checkData(1, "a102",
        "b102Postfix",
        "b102Postfix");

    // send and check same value
    sendData("102");
    checkData(1, "a102",
        "b102Postfix",
        "b102Postfix");

    // send and check new value
    sendData("202");
    checkData(2, "a202",
        "b202Postfix",
        "b202Postfix");

  }

  @Override
  protected void closeConnections() {
    if (handler != null) {
      handler.close();
    }
    if (model != null) {
      model.ragconnectCloseConnections();
    }
  }

  private void sendData(String input) throws InterruptedException {
    publisher.publish(TOPIC_IN, input.getBytes());
    TestUtils.waitForMqtt();
  }

  private void checkData(int expectedNumberOfValues, String expectedLastAValue,
                         String expectedLastB1Value, String expectedLastB2Value) {
    dataA.assertEqualData(expectedNumberOfValues, expectedLastAValue);
    dataB1.assertEqualData(expectedNumberOfValues, expectedLastB1Value);
    dataB2.assertEqualData(expectedNumberOfValues, expectedLastB2Value);
  }

  private static class ReceiverData {
    String lastStringValue;
    int numberOfStringValues = 0;

    public void assertEqualData(int expectedNumberOfValues, String expectedLastValue) {
      assertEquals(expectedNumberOfValues, this.numberOfStringValues);
      assertEquals(expectedLastValue, this.lastStringValue);
    }
  }
}
